//On héritera cette classe pour produire la forme
class drawableBoid
{
  drawableBoid() {
  }
  void draw() {
  }
}


// The Boid class

class Boid {

  PVector location;
  PVector velocity;
  PVector acceleration;
  float r;
  float maxforce;    // Maximum steering force
  float maxspeed;    // Maximum speed
  drawableBoid drawable;
  color couleur;
  float aleat;



  Boid(float x, float y, drawableBoid t, float _speed, float _force)
  {
    drawable = t;
    couleur = color(0,0,0);

    acceleration = new PVector(0, 0);

    // This is a new PVector method not yet implemented in JS
    // velocity = PVector.random2D();

    // Leaving the code temporarily this way so that this example runs in JS
    float angle = random(TWO_PI);
    velocity = new PVector(cos(angle), sin(angle));

    location = new PVector(x, y);
    r = 20.0;
    maxspeed = _speed;
    maxforce = _force;
    aleat = random(1000);
  }



  void run(ArrayList<Boid> boids, ArrayList<Boid> antiGrav, ArrayList<PVector> antiGravPoints)
  {
    flock(boids, antiGrav, antiGravPoints);
    update();
    borders();
    render();
  }



  void applyForce(PVector force)
  {
    // We could add mass here if we want A = F / M
    acceleration.add(force);
  }



  // We accumulate a new acceleration each time based on three rules
  void flock(ArrayList<Boid> boids, ArrayList<Boid> antiGrav, ArrayList<PVector> antiGravPoints)
  {
    PVector[] sepAliCoh = proche(boids, antiGrav, antiGravPoints);   // Separation
    // Arbitrarily weight these forces
    sepAliCoh[0].mult(1.5);
    sepAliCoh[1].mult(1.0);
    sepAliCoh[2].mult(1.0);
    // Add the force vectors to acceleration
    applyForce(sepAliCoh[0]);
    applyForce(sepAliCoh[1]);
    applyForce(sepAliCoh[2]);
  }



  // Method to update location
  void update() {
    // Update velocity
    velocity.add(acceleration);
    // Limit speed
    velocity.limit(maxspeed);
    location.add(velocity);
    // Reset accelertion to 0 each cycle
    acceleration.mult(0);
  }



  // A method that calculates and applies a steering force towards a target
  // STEER = DESIRED MINUS VELOCITY
  PVector seek(PVector target)
  {
    PVector desired = PVector.sub(target, location);  // A vector pointing from the location to the target
    // Scale to maximum speed
    desired.normalize();
    desired.mult(maxspeed);

    // Above two lines of code below could be condensed with new PVector setMag() method
    // Not using this method until Processing.js catches up
    // desired.setMag(maxspeed);

    // Steering = Desired minus Velocity
    PVector steer = PVector.sub(desired, velocity);
    steer.limit(maxforce);  // Limit to maximum steering force

      return steer;
  }



  void render()
  {
    // Draw a triangle rotated in the direction of velocity
    float theta = velocity.heading2D() + radians(90);
    // heading2D() above is now heading() but leaving old syntax until Processing.js catches up

    pushMatrix();
    translate(location.x, location.y);
    rotate(theta);
    fill(couleur);
    stroke(couleur);
    drawable.draw();
    popMatrix();
  }



  // Wraparound
  void borders()
  {
    if (location.x < -r) location.x = width+r;
    if (location.y < -r) location.y = height+r;
    if (location.x > width+r) location.x = -r;
    if (location.y > height+r) location.y = -r;
  }



  // Separation
  // Method checks for nearby boids and steers away
  PVector[] proche (ArrayList<Boid> boids, ArrayList<Boid> antiGrav, ArrayList<PVector> antiGravPoints)
  {
    float desiredseparation = 15.0f;
    float impactGravity = 300;
    PVector steer = new PVector(0, 0, 0);
    int count = 0;

    float neighbordist = 80;
    PVector sumAlign = new PVector(0, 0);
    PVector sumCohesion = new PVector(0, 0);   // Start with empty vector to accumulate all locations
    int countCoh = 0;



    // For every boid in the system, check if it's too close
    for (Boid other : boids) {
      float d = PVector.dist(location, other.location);
      // If the distance is greater than 0 and less than an arbitrary amount (0 when you are yourself)
      if ((d > 0) && (d < desiredseparation)) {
        // Calculate vector pointing away from neighbor
        PVector diff = PVector.sub(location, other.location);
        diff.normalize();
        diff.div(d);        // Weight by distance
        steer.add(diff);
        count++;            // Keep track of how many
      }

      if ((d > 0) && (d < neighbordist))
      {
        sumAlign.add(other.velocity);
        sumCohesion.add(other.location); // Add location
        countCoh++;
      }
    }
    for (Boid gravit : antiGrav) {
      float d = PVector.dist(location, gravit.location);
      // If the distance is greater than 0 and less than an arbitrary amount (0 when you are yourself)
      if ((d > 0) && (d < 2*desiredseparation)) {
        // Calculate vector pointing away from neighbor
        PVector diff = PVector.sub(location, gravit.location);
        diff.normalize();
        diff.div(d/impactGravity);        // Weight by distance
        steer.add(diff);
      }
    }
    for (PVector gravit : antiGravPoints) {
      float d = PVector.dist(location, gravit);
      // If the distance is greater than 0 and less than an arbitrary amount (0 when you are yourself)
      if ((d > 0) && (d < 10*desiredseparation)) {
        // Calculate vector pointing away from neighbor
        PVector diff = PVector.sub(location, gravit);
        diff.normalize();
        diff.div(d/(2*impactGravity));        // Weight by distance
        steer.add(diff);
      }
    }



    // Average -- divide by how many
    if (count > 0)
      steer.div((float)count);
    // As long as the vector is greater than 0
    if (steer.mag() > 0)
    {
      // First two lines of code below could be condensed with new PVector setMag() method
      // Not using this method until Processing.js catches up
      // steer.setMag(maxspeed);

      // Implement Reynolds: Steering = Desired - Velocity
      steer.normalize();
      steer.mult(maxspeed);
      steer.sub(velocity);
      steer.limit(maxforce);
    }



    PVector steerAlign;
    if (countCoh > 0)
    {
      sumCohesion.div((float)countCoh);
      //return seek(sumCohesion);
      sumAlign.div((float)countCoh);
      // First two lines of code below could be condensed with new PVector setMag() method
      // Not using this method until Processing.js catches up
      // sum.setMag(maxspeed);

      // Implement Reynolds: Steering = Desired - Velocity
      sumAlign.normalize();
      sumAlign.mult(maxspeed);
      steerAlign = PVector.sub(sumAlign, velocity);
      steerAlign.limit(maxforce);
    } else
    {
      sumCohesion.mult(0);
      steerAlign = new PVector(0, 0);
    }

    PVector[] res = {
      steer, steerAlign, seek(sumCohesion)
    }; 
    return res;
  }



  void aligner() {
    location.x += ((height*red(couleur)/255)+20 - location.x) * 0.1;
    //location.y += 2*(noise(frameCount*0.01 + aleat)-0.5);
    location.y +=3;
    
    borders();
    pushMatrix();
    translate(location.x, location.y);
    fill(couleur);
    stroke(couleur);
    drawable.draw();
    popMatrix();
  }
  
  void setCouleur(float freq) {
    couleur = color(255-(freq-HAUTEUR_MIN)*255/(HAUTEUR_MAX-HAUTEUR_MIN), 0, (freq-HAUTEUR_MIN)*255/(HAUTEUR_MAX-HAUTEUR_MIN));
  }
  
  void invCouleur() {
    float tmp = blue(couleur)/255;
    couleur = color(tmp*255, 255-tmp*128, 0);
  }
  
  void retourCouleur() {
    float tmp = red(couleur)/255;
    couleur = color(255-tmp*255, 0, tmp*255);
  }
}

